"""
Copyright (c) 2017-2021, Niklas Löcherer, Hendrik Noeller, Professorship for Audio Information Processing, Technical University of Munich, Germany. 

This work is licensed under a Creative Commons Attribution 4.0 International (CC BY 4.0) License, see https://creativecommons.org/licenses/by/4.0/ for more information. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""
"""
Created on Tue Feb 19 20:13:15 2019

@author: Niklas Löcherer

Heavily Edited by Hendrik Noeller on Tue 16 Jun 2020
First I cleaned up the code, put everything relevant
into the arguments and wrote a more robust timing
mechanism including an immediately printed report.
Then I also added manual mode that allows the user
to send arbitrary positions by typing them into
the terminal 
"""

import argparse
import math
import re
import numpy as np
import time
import ctypes

from pythonosc import udp_client

from source_trajectory_line import source_trajectory_line
from source_trajectory_circle import source_trajectory_circle

winmm = ctypes.WinDLL('winmm')
winmm.timeBeginPeriod(1)

if __name__ == "__main__":
    # Open UDP Configuration
    parser = argparse.ArgumentParser()
    parser.add_argument("--rtsofe-ip", default='129.187.247.20')
    parser.add_argument("--rtsofe-port", type=int, default=40000)
    parser.add_argument("--rtsofe-ip-2", default='129.187.247.183')
    parser.add_argument("--rtsofe-port-2", type=int, default=40001)
    parser.add_argument("--rtsofe-ip-3", default='127.0.0.1')
    parser.add_argument("--rtsofe-port-3", type=int, default=40002)
    parser.add_argument("--timefile", default='')
    parser.add_argument("--rate", type=int, default=350)
    parser.add_argument("--trajectory", choices=['line', 'circle'], default='circle')
    parser.add_argument("--point-a", type=float, nargs='+', default = [ 0, 0, 2])
    parser.add_argument("--point-b", type=float, nargs='+', default = [10, 0, 2])
    parser.add_argument("--radius", type=float, default=1)
    parser.add_argument("--revolutions", type=float, default=10)
    parser.add_argument("--velocity", type=float, default=10)
    parser.add_argument("--manual", action='store_true')
    args, unknown = parser.parse_known_args()

    rt_sofe_client = udp_client.SimpleUDPClient(args.rtsofe_ip, args.rtsofe_port)
    rt_sofe_client_2 = udp_client.SimpleUDPClient(args.rtsofe_ip_2, args.rtsofe_port_2)
    rt_sofe_client_3 = udp_client.SimpleUDPClient(args.rtsofe_ip_3, args.rtsofe_port_3)

    
    if args.manual:
        while True:
            position = input("Press enter a position: ")
            if position == 'exit':
                break
            # Match digits with a regex (i wrote this myself it might be buggy)
            position = re.findall(r'(-?\d+(?:[\.,]\d+)?)', position);
            # Replace comma with dot
            position = [s.replace(',','.') for s in position]
            # Convert all the matches (its a list) to floats
            position = list(map(float, position));
            # if there were exactly three numbers send them, else print an error
            
            if len(position)  < 3:
                if len(position) == 0:
                    position.append(0)
                if len(position) == 1:
                    position.append(0)
                    
                print("Converting angle %s" % position)
                radius = args.radius
                az = position[0] / 360 * 2 * math.pi
                el = position[1] / 360 * 2 * math.pi
                x = math.sin(az) * math.cos(el) * radius
                y = math.cos(az) * math.cos(el) * radius
                z = math.sin(el) * radius                   
                position = [x, y, z]
            
            if len(position) == 3:
                print("Sending position %s" % position);
                rt_sofe_client.send_message('/source-pos',position)
                rt_sofe_client_2.send_message('/source-pos',position)
                rt_sofe_client_3.send_message('/source-pos',position)
                
            else:
                print("Please enter three coordinates. Type 'exit' to quit");
    else:
        # initialize source trajectory
        if args.trajectory == 'line':
            d = 1 #minimum distance (m)
            x_s = args.point_a[0] 
            y_s = args.point_a[1] 
            z_s = args.point_a[2] 
            x_e = args.point_b[0] 
            y_e = args.point_b[1] 
            z_e = args.point_b[2]  
            # x,y,z start, x,y,z end,  velocity, fs
            traj = source_trajectory_line(x_s, y_s, z_s, x_e, y_e, z_e, args.velocity, args.rate) # creates line trajectory
        elif args.trajectory == 'circle':
            x_s = args.point_a[0] 
            y_s = args.point_a[1] 
            z_s = args.point_a[2] 
            # center x, y, z   radius, start_angle, end_angle, v, fs
            traj = source_trajectory_circle(x_s, y_s, z_s, args.radius, 0, 3.1415 * 2 * args.revolutions, args.velocity, args.rate) # creates circle trajectory
            
        
        # Send the Data!
        print("Starting to send positions")
        
        # Send the first Position
        rt_sofe_client.send_message('/source-pos',map(float, traj[0]))
        last_sent_index = 0
        times = []
        
        start = time.perf_counter()
        while True:
            # Get the current sample index
            now = time.perf_counter()
            index = round((now - start) * args.rate)
            
            # Continue if we're still in the same block
            if (index == last_sent_index):
                continue
                
            # Stop if the trajectory is over
            if (index >= len(traj)):
                break
              
            rt_sofe_client.send_message('/source-pos',map(float, traj[index]))
            rt_sofe_client_2.send_message('/source-pos',map(float, traj[index]))
            rt_sofe_client_3.send_message('/source-pos',map(float, traj[index]))
            times.append([index, now])
            last_sent_index = index
            time.sleep(0.001) # no one likes busy waiting but python can't sleep at millisecond accuracy.
        
        # Send the last Position just to make sure
        rt_sofe_client.send_message('/source-pos',map(float, traj[-1]))
        rt_sofe_client_2.send_message('/source-pos',map(float, traj[-1]))
        rt_sofe_client_3.send_message('/source-pos',map(float, traj[-1]))


        print("Done sending")
        
        duration = times[-1][1] - times[0][1]
        num_positions = len(times) + 1 # last Position resent
        print("Sent %.2f positions in %.2f seconds (%.2f Hz)" % (num_positions, duration, num_positions / duration))
        
        intervals = [times[i+1][1] - times[i][1] for i in range(0,len(times)-1)]
        avg_interval = np.mean(intervals)
        std_interval = np.std(intervals) 
        max_deviation = max([abs(interval - avg_interval) for interval in intervals])
        print("Average interval was %.2f ms, stddev = %.2f ms, maximum deviation = %.2f ms" % (avg_interval * 1000, std_interval * 1000, max_deviation * 1000))

        if len(args.timefile):
            timefile = open(args.timefile, 'w')
            for (index, time) in times:
                timefile.write("%s\t%s\n" % (index, time))
            timefile.close()
